// SPDX-License-Identifier: MIT

pragma solidity ^0.4.0;
pragma experimental ABIEncoderV2;

contract ExperimentalABIEncoderV2 {
    function foo() pure public returns(bool) {
        return true;
    }
}
